/**
 * @file imgdraw.c
 * @brief Functions to draw into yu'sv images
 * @author Florian Schmitt (0schmitt@informatik.uni-hamburg.de)
 * @date Jul-2012
**/
#include "camcapture.h"
#include <math.h>

/** @cond */
#define PI 3.14159265
/** @endcond */




/**
 * Draws a line via x axis
 * @param image Image data struct
 * @param fromx x coordinate to start from
 * @param fromy y coordinate to start from
 * @param tox   x coordinate to draw to
 * @param toy   y coordinate to draw to
 * @param color to draw with (rgb)
 * @param delta of the line
**/
void camcapture_draw_line_over_x(image_data_t *image, int fromx, int fromy, int tox, int toy, color_rgb_t *color, double delta)
{
	int pos, tmp;
	int Y, Cb, Cr;
	void *data;
	rgb_to_ycbcr(color, &Y, &Cb, &Cr);
	data = image->img;
	
	/* flip direction if needed */
	if(tox < fromx) 
	{
		tmp = tox;
		tox = fromx;
		fromx = tmp;
		tmp = toy;
		toy = fromy;
		fromy = tmp;
	} 
	/* loop thru x */
	for(pos = fromx; pos<=tox; pos++)
	{
		/* check if position is inside the image */
		if(pos < 0) break;
		if(pos >= image->width) break;
		/* calculate y value into tmp */
		tmp = fromy + (int)( delta * (pos-fromx) );
		if((tmp < 0) || (tmp > image->height)) break;
		/* calculate data position into tmp */
		tmp = (tmp*image->width + pos) / 2;
		/* set pixel */
		*((int*)data+tmp) = 0 | (Y) | (Cb << 8) | (Y << 16) | (Cr << 24);
	}
}

/**
 * Draws a line via y axis
 * @param image Image data struct
 * @param fromx x coordinate to start from
 * @param fromy y coordinate to start from
 * @param tox   x coordinate to draw to
 * @param toy   y coordinate to draw to
 * @param color to draw with (rgb)
 * @param delta of the line
**/
void camcapture_draw_line_over_y(image_data_t *image, int fromx, int fromy, int tox, int toy, color_rgb_t *color, double delta)
{
	int pos, tmp;
	int Y, Cb, Cr;
	void *data;
	rgb_to_ycbcr(color, &Y, &Cb, &Cr);
	data = image->img;
	
	/* flip direction if needed */
	if(toy < fromy) 
	{
		tmp = tox;
		tox = fromx;
		fromx = tmp;
		tmp = toy;
		toy = fromy;
		fromy = tmp;
	} 
	/* loop thru x */
	for(pos = fromy; pos<=toy; pos++)
	{
		/* check if position is inside the image */
		if(pos < 0) break;
		if(pos >= image->height) break;
		/* calculate x value into tmp */
		tmp = fromx + (int)( delta * (pos-fromy) );
		if((tmp < 0) || (tmp > image->width)) break;
		/* calculate data position into tmp */
		tmp = (pos*image->width + tmp) / 2;
		/* set pixel */
		*((int*)data+tmp) = 0 | (Y) | (Cb << 8) | (Y << 16) | (Cr << 24);
	}
}


/**
 * Draws a line from one point to another
 * @param image Image data struct
 * @param fromx x coordinate to start from
 * @param fromy y coordinate to start from
 * @param tox   x coordinate to draw to
 * @param toy   y coordinate to draw to
 * @param color to draw with (rgb)
**/
void camcapture_draw_line(image_data_t *image, int fromx, int fromy, int tox, int toy, color_rgb_t *color)
{
	int dx, dy;
	/* check conditions */
	if( (fromx < 0) || (fromx > image->width) ) return;
	if( (fromy < 0) || (fromy > image->height) ) return;
	if( (tox < 0) || (tox > image->width) ) return;
	if( (toy < 0) || (toy > image->height) ) return;
	/* draw */
	dx = deltaval(fromx, tox);
	dy = deltaval(fromy, toy);
	if( absolute(dx) >= absolute(dy) )
	{
		camcapture_draw_line_over_x(image, fromx, fromy, tox, toy, color, (double)dy/dx);
	}
	else
	{
		camcapture_draw_line_over_y(image, fromx, fromy, tox, toy, color, (double)dx/dy);
	}
}




/**
 * Sets a pixel to a given color
 * @param image Image data struct
 * @param px x coordinate
 * @param px y coordinate
 * @param color to set (rgb)
**/
void camcapture_set_pixel(image_data_t *image, int px, int py, color_rgb_t *color)
{
	int tmp, Y, Cb, Cr;
	void *data;

	/* calc color and get image pointer */
	rgb_to_ycbcr(color, &Y, &Cb, &Cr);
	data = image->img;
	
	/* check if position is inside the image */
	if(px < 0) return;
	if(px > image->width) return;
	if(py < 0) return;
	if(py > image->height) return;

	/* calculate data position into tmp */
	tmp = (py*image->width + px) / 2;
	/* set pixel */
	*((int*)data+tmp) = 0 | (Y) | (Cb << 8) | (Y << 16) | (Cr << 24);
}


/**
 * Draws a cross marker
 * @param image Image data struct
 * @param x position
 * @param y position
 * @param color to draw with (rgb)
**/
void camcapture_draw_cross(image_data_t *image, int x, int y, color_rgb_t *color)
{
	camcapture_draw_line(image, x-3, y,   x+4, y,   color);
	camcapture_draw_line(image, x,   y-3, x,   y+3, color);
}


/**
 * Draws a cross marker by a imaeg_point object
 * @param image Image data struct
 * @param ip  image_point object
 * @param color to draw with (rgb)
**/
void camcapture_draw_cross_obj(image_data_t *image, image_point_t *ip, color_rgb_t *color)
{
	if(ip == 0) return;
	camcapture_draw_cross(image, ip->x, ip->y, color);
}


/**
 * Draws a cross marker of many points collected in a list object
 * @param image Image data struct
 * @param list  image_point_list_t object
 * @param color to draw with (rgb)
**/
void camcapture_draw_point_list_crossed(image_data_t *image, image_point_list_t *list, color_rgb_t *color)
{
	if(list == 0) return;
	while((list = list->next) != 0)
	{
		camcapture_draw_cross(image, list->data->x, list->data->y, color);
	}
}


/**
 * Draws rectangles of many points collected in a list object
 * The point will be the rect's center
 * @param image Image data struct
 * @param list  image_point_list_t object
 * @param size  x=y size of the rect
 * @param color to draw with (rgb)
**/
void camcapture_draw_point_list_rects(image_data_t *image, image_point_list_t *list, int size, color_rgb_t *color)
{
	if(list == 0) return;
	while((list = list->next) != 0)
	{
		camcapture_draw_rect(image, 
		                     list->data->x - (size/2), 
		                     list->data->y - (size/2), 
		                     list->data->x + (size/2), 
		                     list->data->y + (size/2), 
		                     color);
	}
}


/**
 * Draws the path of a image_point_list_t
 * @param image Image data struct
 * @param list  image_point_list_t object
 * @param color to draw with (rgb)
**/
void camcapture_draw_point_list_path(image_data_t *image, image_point_list_t *list, color_rgb_t *color)
{
	if(list == 0) return;
	/* get first node */
	image_point_list_t *previous = list->next;
	list = list->next;
	if(list == 0) return;
	/* loop through rest */
	while((list = list->next) != 0)
	{
		camcapture_draw_line(image, previous->data->x, previous->data->y, list->data->x, list->data->y, color);
		previous = list;
	}
}


/**
 * Draws a (not filled) rectangle
 * @param image Image data struct 
 * @param x1 upper left edge x
 * @param y1 upper left edge y
 * @param x2 lower right edge x
 * @param y2 lower right edge y
 * @param color to draw with (rgb)
**/
void camcapture_draw_rect(image_data_t *image, int x1, int y1, int x2, int y2, color_rgb_t *color)
{
	/* top line */
	camcapture_draw_line(image, x1, y1, x2, y1, color);
	/* bottom line */
	camcapture_draw_line(image, x1, y2, x2, y2, color);
	/* left line */
	camcapture_draw_line(image, x1, y1, x1, y2, color);
	/* right line */
	camcapture_draw_line(image, x2, y1, x2, y2, color);
}



/**
 * Draws a filled rectangle
 * @param image Image data struct 
 * @param x1 upper left edge x
 * @param y1 upper left edge y
 * @param x2 lower right edge x
 * @param y2 lower right edge y
 * @param color to draw and fill with (rgb)
**/
void camcapture_draw_fillrect(image_data_t *image, int x1, int y1, int x2, int y2, color_rgb_t *color)
{
	int posx, posy, tmp;
	int Y, Cb, Cr;
	void *data;
	rgb_to_ycbcr(color, &Y, &Cb, &Cr);
	data = image->img;
	
	/* check sizes */
	if( (x1 > x2) || (y1 > y2) ) 
		return;
	/* loop thru x and y */
	for(posy = y1; posy<=y2; posy++)
	{
		for(posx = x1; posx<=x2; posx++)
		{
			/* calculate data position into tmp */
			tmp = (posy*image->width + posx) / 2;
			/* set pixel */
			*((int*)data+tmp) = 0 | (Y) | (Cb << 8) | (Y << 16) | (Cr << 24);
		}
	}
}






/**
 * Draws a circle
 * @param image Image data struct 
 * @param px  x coordinate of center
 * @param py  y coordinate of center
 * @param rad radius of the circle
 * @param color to draw with (rgb)
**/
void camcapture_draw_circle(image_data_t *image, int px, int py, int rad, color_rgb_t *color)
{
	int pos, tmp, tmpx, tmpy, res;
	int Y, Cb, Cr;
	void *data;
	rgb_to_ycbcr(color, &Y, &Cb, &Cr);
	data = image->img;
	
	/* set the circles resolution */
	res = rad * 2 * PI;
	
	/* loop thru x */
	for(pos = 0; pos<=res; pos++)
	{
		/* calculate values */
		tmpx = (int) (sin((float)pos/res*2*PI) * rad) + px;	
		tmpy = (int) (cos((float)pos/res*2*PI) * rad) + py;
		if( (tmpx < 0) || (tmpy < 0) ) continue;
		if( (tmpx > image->width) || (tmpy > image->height) ) continue;
		/* calculate data position into tmp */
		tmp = (tmpy*image->width + tmpx) / 2;
		/* set pixel */
		*((int*)data+tmp) = 0 | (Y) | (Cb << 8) | (Y << 16) | (Cr << 24);
	}
}



/**
 * Draws a circle from a circle object
 * @param image Image data struct 
 * @param circle the circle object
 * @param color to draw with (rgb)
**/
void camcapture_draw_circle_obj(image_data_t *image, image_circle_t *circle, color_rgb_t *color)
{
	if(circle == 0) return;
	camcapture_draw_circle(image, circle->x, circle->y, circle->rad, color);
}
