/**
 * @file listsort.c
 * @brief Provides functions for sorting lists
 * @author Florian Schmitt (0schmitt@informatik.uni-hamburg.de)
 * @date Aug-2012
**/
#include "camcapture.h"
#include <math.h>
#include <stdlib.h>

/** @cond */
#define PI 3.14159265
/** @endcond */





/**
 * Get and remove first point from list
 * @param list    The image point list 
 * @param element The element to add
**/
void addTo(image_point_list_t *list, image_point_list_t *element)
{
	image_point_list_t *old = list->next;
	list->next = element;
	element->next = old;
}



/**
 * Get the nearest point to point if it is in maxdist
 * @param list  The image point list to search in
 * @param point The image point to look around him
 * @param rad   Maximum distance that the point may be away from this one (if 0 all points)
**/
image_point_list_t *getAndRemoveNextPointTo(image_point_list_t *list, image_point_t *point, int maxdist)
{
	if(list == 0) return 0;
	if(list->next == 0) return 0;
	if(point == 0) return 0;
	image_point_list_t *next;
	int nextdist;
	image_point_list_t *akt = list->next;
	image_point_list_t *beforenext = list;
	image_point_list_t *beforeakt = list;
	/* first value */
	next = akt;
	nextdist = camcapture_math_distance(point, next->data);
	akt = akt->next;
	while(akt != 0)
	{
		if(akt->data == 0) continue;
		int aktdist = camcapture_math_distance(point, akt->data);
		/* this is the next point - set its parameters */
		if(aktdist < nextdist)
		{
			beforenext = beforeakt;
			next = akt;
			nextdist = aktdist;
		}
		beforeakt = akt;
		akt = akt->next;
	}
	if( (maxdist != 0) && ((float)nextdist > (float)maxdist) ) return 0;
	/* now remove point from old list */
	beforenext->next = next->next;
	return next;
}



/**
 * Get and remove first point from list
 * @param list The image point list 
**/
image_point_list_t *getAndRemoveFirst(image_point_list_t *list)
{
	if(list == 0) return 0;
	if(list->next == 0) return 0;
	image_point_list_t *next = list->next;
	list->next = next->next;
	next->next = 0;
	return next;
}





/**
 * Sort the list by going through a circle
 * @param list   The image point list reference
 * @param center The point lists central point
 * @param rad    The radius of the circle to detect
**/
void camcapture_point_list_sort_by_circle(image_point_list_t *list, image_point_t *center, int rad)
{
	/* check failures */
	if(list == 0) return;
	if(list->next == 0) return;
	if(center == 0) return;
	int res, pos;
	/* save startup node */
	image_point_list_t *newlist = camcapture_create_point_list();
	/* go in a circle around */
	res = (float) rad * 2 * PI;
	for(pos=0; pos<res; pos++)
	{
		image_point_t p;
		p.x = (int)(sin((float)pos/res*2*PI) * rad) + center->x;
		p.y = (int)(cos((float)pos/res*2*PI) * rad) + center->y;
		if(list->next == 0) break;
		/* get and remove point from list */
		image_point_list_t *add = getAndRemoveNextPointTo(list, &p, (float)rad*0.225f);
		if(add != 0) 
		{
			/* insert at beginning of newlist */
			image_point_list_t *old = newlist->next;
			newlist->next = add;
			add->next = old;
		}
	}
	/* cleanup rest from old list*/
	image_point_list_t *next = list->next;
	image_point_list_t *former;
	while(next != 0)
	{
		former = next;
		next = next->next;
		if(former != 0) free(former);
	} 
	/* cleanup newlist startup node */
	list->next = newlist->next;
	free(newlist);
}





/**
 * Sort the list by going getting the next one (insertion sort)
 * @param list The image point list reference
**/
void camcapture_point_list_sort_by_nextto(image_point_list_t *list)
{
	/* check failures */
	if(list == 0) return;
	if(list->next == 0) return;
	/* save startup node */
	image_point_list_t *newlist = camcapture_create_point_list();
	image_point_list_t *akt = getAndRemoveFirst(list);
	/* go through old list */
	while(akt != 0)
	{
		/* add akt to new list */
		akt->next = newlist->next;
		newlist->next = akt;
		/* get next */
		akt = getAndRemoveNextPointTo(list, akt->data, 0);
	}
	/* cleanup newlist startup node */
	list->next = newlist->next;
	free(newlist);
}
